
/* prob31.c						*/
/* program for LCA31 option 't'				*/
/* calculate probabilities related to evolution		*/
/* Harold V. McIntosh, 10 August 1987			*/
/* 24 March 1989 - Bernstein polynomial evaluation	*/
/* 15 September 1989 - entropy calculations		*/
/* 29 November 1989 - graph new lambda vs old lambda	*/

/* references:							*/
/*								*/
/* W. John Wilbur, David J. Lipman and Shihab A. Shamma		*/
/* On the prediction of local patterns in cellular automata	*/
/* Physica 19D 397-410 (1986)					*/ 
/*								*/
/* Howard A. Gutowitz, Jonathan D. Victor and Bruce W. Knight	*/
/* Local structure theory for cellular automata			*/
/* Physica 28D 18-48 (1987)					*/

/*	Copyright (C) 1987	*/
/*	Copyright (C) 1988	*/
/*	Copyright (C) 1989	*/
/*	Harold V. McIntosh	*/
/*	Gerardo Cisneros S.	*/

# include <math.h>
# include <lcau31.h>

extern char   ascrule[KK][KK][KK];		/* ASCII transition values */
extern int    binrule[KK][KK][KK];		/* binary transition values */
extern int    arr1[AL];				/* line of cells */

# define BER	    0	/* row (from bottom) for Bernstein p.	*/
# define BEC	  205	/* column for Bernstein Polynomial	*/
# define BROW	   13   /* row for bar charts			*/
# define EROW	   22   /* row for evolution synopsis		*/
# define TROW      89   /* row for t-triangle			*/
# define TCOL     205   /* column for t-triangle		*/

/* edit the probability screen	*/
edtri() {char c;
  videomode(COLGRAF);
  videopalette(YELREGR);
  videotrgrid(TROW,TCOL,10,3);

  while (0<1) {
  videocursor(0,0,0); scrrul();
  videocursor(0,0,36);
  videoputc('?',2);
  c=kbdin();
  if (c == '\015') break;
  videocursor(0,0,38);
  videoputc(c,2);
  videocursor(0,0,36);
  videoputc(' ',0);
  switch (c) {
    case '+': videopalette(WHCYMAG); break;
    case '-': videopalette(YELREGR); break;
    case 'a': nblclr(); asfreq(TROW,TCOL,3); break;
    case 'b': bpanel(); break;
    case 'e': pevolve(); break;
    case 'g': lifreq(50,TROW,TCOL,2); break;
    case 'G': lifreq(200,TROW,TCOL,1); break;
    case 'h': penry(37);  break;
    case 'H': penry(100); break;
    case 'l': lpanel(); break;
    case 'm': nblclr(); moncar(1,2); break;
    case 'M': nblclr(); moncar(50,1); break;
    case 'n': nblclr(); neisum(8*BROW,AL,48); break;
    case 'N': nblclr(); nbrsum(8*BROW,AL,48); break;
    case 's': spdiff(50,6); break;
    case 't': pdiff(39,6);  break;
    case 'T': pdiff(100,6); break;
    case 'x': ranprul(); videocursor(0,0,0); scrrul(); break;
    case 'X': ranrul(); videocursor(0,0,0); scrrul(); break;
    case '1': nblclr(); oneblfreq(8*BROW,300,48); break;
    case '2': nblclr(); twoblfreq(8*BROW,300,48); break;
    case '3': nblclr(); thrblfreq(8*BROW,300,48); break;
    case '4': nblclr(); foublfreq(8*BROW,300,48); break;
    case '5': nblclr(); fivblfreq(8*BROW,300,48); break;
    case 'z': nblclr(); break;
    case '/': videomode(COLGRAF); videopalette(YELREGR); break;
    case '?': nblclr(); trmenu(); break;
    default: break;
    } /* end switch */
  }   /* end while  */
  videopalette(WHCYMAG);
  videomode(T80X25);
}      /* end edtri  */

/* show menu */
trmenu() {
  videocursor(0,BROW,0);
  printf("a,h,H  - a priori estimates, entropy\n");
  printf("m,M,g,G  sample evolution\n");
  printf("s      - graph probabilities in stereo\n");
  printf("t,T    - graph mean field probabilities\n");
  printf("12345  - n-block bar charts\n");
  printf("+-     - change color pallette\n");
  printf("e      - 16 lines evolution\n");
  printf("z/?    - clear panel, screen; show menu\n");
  printf("<cr>   - exit\n");
}

/* show sixteen lines of evolution on screen */
pevolve() {int i, j;
  videoscroll(EROW,0,EROW+1,40,0,0);
  asctobin();
  for (j=8*EROW; j<8*EROW+16; j++) {
    for (i=0; i<AL; i++) videodot(j,i,arr1[i]);
    onegen(AL);
    }
}

/* Clear a space for the n-block frequencies */
nblclr() {videoscroll(BROW,0,BROW+8,40,0,0);}

/* make a frame for a graph 		 */
/* (x,y) = lower left corner; e.g. (0,0) */
/* n     = dimension of frame            */
gfram(x,y,n) int x, y, n; {int i;

for (i=0; i<=n; i++) videodot(199-y-i,x,0);
for (i=0; i<=n; i++) videodot(199-y-i,x+n,0);
for (i=0; i<=n; i++) videodot(199-n-y,x+i,0);
for (i=0; i<=n; i++) videodot(199-y,x+i,0);

for (i=0; i<=n; i+=10) videodot(199-y-i,x,3);
for (i=0; i<=n; i+=10) videodot(199-y-i,x+n,3);
for (i=0; i<=n; i+=10) videodot(199-n-y,x+i,3);
for (i=0; i<=n; i+=10) videodot(199-y,x+i,3);
}

/* display state frequencies in arr1 as a dot in triangle at (u,v) */
/* n = generations of evolution					   */
/* (u,v) = corner of triangle					   */
/* l = color of dot						   */

lifreq(n,u,v,l) int n, u, v, l; {
int i, ii;
int stat[KK];
double staf[KK], s;

s=1.0/((double)(AL));
asctobin();
for (ii=0; ii<n; ii++) {
  onegen(AL);
  for (i=0; i<KK; i++) stat[i]=0;
  for (i=0; i<AL; i++) (stat[arr1[i]])++;
  for (i=0; i<KK; i++) staf[i]=s*((double)stat[i]);
  videotrdot(u,v,staf,l);
  }
}

/* bdot(i,j,l) - place dot on Bernstein screen at (i,j) in color l */
bdot(i,j,l) int i, j, l; {videodot(199-BER-j,BEC+i,l);}

/* pdot(i,j,l) - place polkadot on Bernstein screen at (i,j) in color l */
pdot(i,j,l) int i, j, l; {videodot(199-BER-j,BEC+i,l+i%2);}

/* plot a single point on a triangular grid */

videotrdot(u,v,t,l) double t[KK]; int u, v, l; {double s, y, z;
s=199.0/(t[0]+t[1]+t[2]); y=t[1]*s; z=t[2]*s;
videodot(u-(int)(0.433*z),v+(int)(0.250*(y+y+z)),l);
}

/* plot a stereo point on a triangular grid */
/* (u,v) screen location lower left corner  */
/* (x,y,z) triangular coordinates	    */
/* t       height			    */
/* l       color			    */

videostdot(u,v,x,y,z,t,l) double x, y, z, t; int u, v, l; {double s;
s=199.0/(x+y+z); x=x*s; y=y*s; z=z*s;
videodot(u-(int)(0.433*z),v+(int)(0.250*(y+y+z+t)),l);
}

/* set up a triangular gridwork */
/* (u,v) = corner of triangle	*/
/* n = number of rulings	*/
/* l = color of lines		*/

videotrgrid(u,v,n,l) int u, v, n, l; {
int i, j;
double a[KK], en;
if(n==0) return;
en=(double)n;
for (i=0; i<=n;   i++) {
for (j=0; i+j<=n; j++) {
  a[0]=((double)(i)/en);
  a[1]=((double)(j)/en);
  a[2]=1.0-a[0]-a[1];
  videotrdot(u,v,a,l);
  }}
}

/* options 'm', 'M'.				*/
/* "Monte Carlo" determination of probabilities */
/* n = mumber of generations			*/
/* l = color of dot				*/

moncar(n,l) int n, l; {
int i, j, k, b[KK], bb[KK][KK];
double bf[KK], bbf[KK][KK];
double a[KK], aa[KK], d[KK], el, en, t, tt, s;

nblclr();
asctobin();
el=1.0/((double)AL);
en=1.0/((double)n);
for (i=0; i<KK; i++) {a[i]=0.0; aa[i]=0.0;}
for (k=0; k<n; k++) {
  onegen(AL);
  for (i=0; i<KK; i++) b[i]=0;
  for (i=0; i<AL; i++) b[arr1[i]]++;
  for (i=0; i<KK; i++) bf[i]=((double)(b[i]))*el;
  for (i=0; i<KK; i++) {a[i]+=bf[i]; aa[i]+=bf[i]*bf[i];}
  for (i=0; i<KK; i++) for (j=0; j<KK; j++) bb[i][j]=0;
  for (i=1; i<AL; i++) bb[arr1[i-1]][arr1[i]]++;
  bb[arr1[AL-1]][arr1[0]]++; 
  for (i=0; i<KK; i++) for (j=0; j<KK; j++) 
    bbf[i][j]=((double)(bb[i][j]))*el;
  videotrdot(TROW,TCOL,bf,l);
  }
for (i=0; i<KK; i++) {a[i]*=en; aa[i]*=en;}
videocursor(0,BROW+1,0);
s=1.0/log((double)KK);
t=0.0; tt=0.0;
for (i=0; i<KK; i++) t+=bf[i]*log(bf[i]);
for (i=0; i<KK; i++) d[i]=sqrt(aa[i]-a[i]*a[i]);
for (i=0; i<KK; i++) for (j=0; j<KK; j++) tt+=bbf[i][j]*log(bbf[i][j]);
printf("lambda       = %5.3f\n",1.0-bf[0]);
printf("entropy      = %5.3f, pairs  %5.3f\n",-t*s,-0.5*tt*s);
printf("\n");
printf("frequency    = %5.3f, %5.3f, %5.3f\n",a[0],a[1],a[2]);
if (n>1) printf("deviation    = %5.3f, %5.3f, %5.3f\n",d[0],d[1],d[2]);
printf("pair freq    = %5.3f, %5.3f, %5.3f\n",bbf[0][0],bbf[0][1],bbf[0][2]);
printf("             = %5.3f, %5.3f, %5.3f\n",bbf[1][0],bbf[1][1],bbf[1][2]);
printf("             = %5.3f, %5.3f, %5.3f\n",bbf[2][0],bbf[2][1],bbf[2][2]);
}

/* option 'n'						*/
/* empirical distribution of sums within a neighborhood */
/* ll - initial line	*/
/* mm - number of cells	*/
/* nn - number of lines	*/

neisum(ll,mm,nn) int ll, mm, nn; {
int ii, i, k, m, n, l, b[TS];
double em, en, bf[TS], bft[TS];

asctobin();
em=((double)(mm));
for (i=0; i<TS; i++) bft[i]=0.0;
k=0;
for (ii=0; ii<nn; ii++) {
  onegen(mm);
  onesum(mm,TS,b);
  for (i=0; i<TS; i++) bf[i]=((double)(b[i]));
  for (i=0; i<TS; i++) bft[i]+=bf[i];
  m=0; videodot(ll+ii,m++,3);
  for (i=0; i<TS; i++) {
    n=(int)(bf[i]);
    if (n>0) for (l=0; l<n; l++) videodot(ll+ii,m++,i%2);
    }
  videodot(ll+ii,m++,3);
  k++;
  if (kbdst()) {kbdin(); break;}
  }  /* end for-ii */
  en=((double)(k));
  m=0; videodot(ll+nn,m++,3);
  for (i=0; i<TS; i++) {
    n=(int)(bft[i]/en);
    if (n>0) for (l=0; l<n; l++) videodot(ll+nn,m++,i%2+1);
    }
  videodot(ll+nn,m++,3);
for (i=0; i<TS; i++) bft[i]/=en*em;
videocursor(0,BROW+7,0);
for (i=0; i<TS; i++) printf("%1d:%5.3f ",i,bft[i]);
/* videocursor(0,BROW+8,0);				*/
/* for (i=5; i<TS; i++) printf("%1d:%5.3f ",i,bft[i]);	*/
}

/* option 'N' */
/* empirical distribution of sums within a neighborhood */
/* ll - initial line	*/
/* mm - number of cells	*/
/* nn - number of lines	*/

nbrsum(ll,mm,nn) int ll, mm, nn; {
int ii, jj, i, k, m, n, l, b[TS];
double em, en, kk, bf[TS], bft[TS];

asctobin();
em=((double)(mm));
en=((double)(nn));
for (i=0; i<TS; i++) bft[i]=0.0;
k=0;
for (jj=0; jj<nn; jj++) {
  for (i=0; i<TS; i++) bf[i]=0.0;
  for (ii=0; ii<nn; ii++) {
    onegen(mm);
    onesum(mm,TS,b);
    for (i=0; i<TS; i++) bf[i]+=((double)(b[i]));
    }  /* end for-ii */
  m=0; videodot(ll+jj,m++,3);
  for (i=0; i<TS; i++) {
    n=(int)(bf[i]/en);
    if (n>0) for (l=0; l<n; l++) videodot(ll+jj,m++,i%2+1);
    }
  videodot(ll+jj,m++,3);
  for (i=0; i<TS; i++) bft[i]+=bf[i];
  k++;
  if (kbdst()) {kbdin(); break;}
  } /* end for-jj */
kk=((double)(k));
videodot(ll+jj,m++,3);
for (i=0; i<TS; i++) bft[i]/=kk*en*em;
videocursor(0,BROW+7,0);
for (i=0; i<TS; i++) printf("%1d:%5.3f ",i,bft[i]);
/* videocursor(0,BROW+8,0);				*/
/* for (i=5; i<TS; i++) printf("%1d:%5.3f ",i,bft[i]);	*/
}

/* gather totalistic neighborhood sums for one generation */
/* mm = number of cells */
/* b[cc]  = sum counts  */

onesum(mm,cc,b) int mm, cc, *b; {int i, j, k;
  for (i=0; i<cc; i++) b[i]=0;
  k=arr1[mm-1]+arr1[0]+arr1[1]; b[k]++;
  for (i=1; i<mm-2; i++) {k=0; for (j=i-1; j<i+2; j++) k+=arr1[j]; b[k]++;}
  k=arr1[mm-2]+arr1[mm-1]+arr1[0]; b[k]++;
}

/* make a grid for the fixpoint graph  */
/* n     = dimension of grid           */
ggrid(n) int n; {int i, j;
for (i=0; i<=n; i+=10) for (j=0; j<=n; j+=10) bdot(i,j,3);}

/* put a diagonal in the fixpoint graph */
gdiag(n) int n; {int i; for (i=0; 2*i<n; i++) bdot(2*i,2*i,3);}

/* plot self-consistent probability graph */
bpanel()  {ggrid(100); gdiag(100); bgraf(100);}

/* plot self-consistent graph in Langton's parameter */
lpanel()  {ggrid(100); gdiag(100); lgraf(100);}

/* graph Bernstein polynomial */
bgraf(n) int n; {
int    i;
double bern(), el, em, en, p, q, t[KK];
en=(double)(n); em=1.0/en; el=1.0/((double)(KK-1));
for (p=0.0, i=0; i<n; p+=em, i++) {
  q=el*(1.0-p); t[1]=t[2]=q;
  t[0]=p; bdot(i,(int)(en*bern(0,t)),3);
  t[0]=q; t[1]=p; bdot(i,(int)(en*bern(1,t)),1);
  t[1]=q; t[2]=p; bdot(i,(int)(en*bern(2,t)),2);
  }
}

/* graph Langton's parameter */
lgraf(n) int n; {int i;
double bern(), el, em, en, p, t[KK];
en=(double)(n); em=1.0/en; el=1.0/((double)(KK-1));
for (p=0.0, i=0; i<n; p+=em, i++) {
  t[0]=p; t[1]=t[2]=el*(1.0-p);
  pdot(n-i,n-(int)(en*bern(0,t)),1);
  }
}

/* evaluate the Bernstein polynomial for component i at point t[KK] */
double bern(i,t) int i; double t[KK]; {int i0, i1, i2; double a, b, c;
asctobin();
for (i0=0, a=0.0; i0<KK; i0++) {
  for (i1=0, b=0.0; i1<KK; i1++) {
    for (i2=0, c=0.0; i2<KK; i2++)
      if (binrule[i0][i1][i2]==i) c+=t[i2];
    b+=c*t[i1];}
  a+=b*t[i0];}
return a;
}

/* graph the entropy over a triangle */

penry(n) int n; {
int    i, j, lnco();
double a[KK], t, em, en;
if (n==0) return;
em=1.0/log(3.0);
en=1.0/((double)(n));
for (i=0; i<=n; i++) {
for (j=0; i+j<=n; j++) {
  t=0.0;
  a[0]=((double)(i))*en; if (a[0]>0.005) t+=a[0]*log(a[0]);
  a[1]=((double)(j))*en; if (a[1]>0.005) t+=a[1]*log(a[1]);
  a[2]=1.0-a[0]-a[1]; if (a[2]>0.005) t+=a[2]*log(a[2]);
  videotrdot(TROW,TCOL,a,lnco(-t*em,i));
  if (kbdst()) {kbdin(); return;}
  }}
}

/* graph the probability differential for state l over a triangle */

pdiff(n,l) int n, l; {
int    i, j, sqco();
double a[KK], en, t, sqsu();
if (n==0) return;
if (l!=6) return;
en=1.0/((double)(n));
for (i=0; i<=n; i++) {
for (j=0; i+j<=n; j++) {
  a[0]=((double)(i))*en;
  a[1]=((double)(j))*en;
  a[2]=1.0-a[0]-a[1];
  t=sqsu(a);
  videotrdot(TROW,TCOL,a,sqco(t,i));
  if (kbdst()) {kbdin(); return;}
  }}
}

/* stereo graph of probability differential for state l over a triangle */

spdiff(n,l) int n, l; {int i, j; double a, b, c, en, t, sqsu();
if (n==0) return;
if (l!=6) return;
en=1.0/((double)(n));
for (i=0; i<=n; i++) {
for (j=0; i+j<=n; j++) {
  a=((double)(i))*en;
  b=((double)(j))*en;
  c=1.0-a-b;
  t=sqsu(a,b,c);
  videostdot(99,5,a,b,c,-12.0*t,sqco(t,i));
  videostdot(99,105,a,b,c,12.0*t,sqco(t,i));
  if (kbdst()) {kbdin(); return;}
  }}
}

/* calculate the squaresum of the three polynomials at a point */

double sqsu(t) double t[KK]; {double u, v, w, bern();
u=bern(0,t)-t[0];
v=bern(1,t)-t[1];
w=bern(2,t)-t[2];
return (u*u+v*v+w*w);
}

/* generate linearly compatible contour values */

int lnco(t,i) int i; double t; {int l;
if (t<0.125) l=0; else {
if (t<0.250) {if (i%2) l=0; else l=3;} else {
if (t<0.375) l=3; else {
if (t<0.500) {if (i%2) l=3; else l=2;} else {
if (t<0.625) l=2; else {
if (t<0.750) {if (i%2) l=2; else l=1;} else {
if (t<0.875) l=1; else {
	      if (i%2) l=1; else l=0;
   }}}}}}}
return l;
}

/* generate squaresum compatible contour values */

int sqco(t,i) int i; double t; {int l;
if (t<0.0075) l=0; else {
if (t<0.0750) {if (i%2) l=0; else l=3;} else {
if (t<0.1250) l=3; else {
if (t<0.2500) {if (i%2) l=3; else l=2;} else {
if (t<0.5000) l=2; else {
if (t<0.7500) {if (i%2) l=2; else l=1;} else {
if (t<1.0000) l=1; else {
	      if (i%2) l=1; else l=0;
   }}}}}}}
return l;
}

/* graph the frequencies of ascrule in color l	 */
/* put dot at coordinates (u,v)			 */
/* percentages evolving into each value		 */
/* percentages shifting - adds to more than 100% */
/* entropy and Langton's lambda = % nonquiescent */

asfreq(u,v,l) int u, v, l; {
int    i, j, i0, i1, i2;
int    stat[KK], stal, stac, star;		/* statistic counts */
double staf[KK], t, pp;

pp=1.0/((double)(KK*KK*KK));
stal=0; stac=0; star=0;
for (i=0; i<KK; i++) stat[i]=0;
for (i0=0; i0<KK; i0++) {
for (i1=0; i1<KK; i1++) {
for (i2=0; i2<KK; i2++) {
  j=ascrule[i0][i1][i2]-'0';
  stat[j]++;
  if(j==i0) star++;
  if(j==i1) stac++;
  if(j==i2) stal++;
  }}}
videocursor(0,BROW,0);
for (i=0; i<KK; i++) staf[i]=((double)stat[i])*pp;
for (i=0; i<KK; i++) printf("%d freq - %5.2f\n",i,staf[i]);
printf("\n");
printf("left   - %5.2f\n",((double)stal)*pp);
printf("still  - %5.2f\n",((double)stac)*pp);
printf("right  - %5.2f\n",((double)star)*pp);
videotrgrid(u,v,10,3);
videotrdot(u,v,staf,l);
for (i=0, t=0.0; i<KK; i++) t+=staf[i]*log(staf[i]);
videocursor(0,BROW,20);
printf("lambda:   %5.3f",1.0-staf[0]);
videocursor(0,BROW+1,20);
printf("entropy:  %5.3f",-t/log((double)KK));
videocursor(0,BROW,0);
}

/* evaluate the one-block probabilities after one generation */

onebl(x,a) double x[KK], a[KK]; {int i0, i1, i2, j0;

for (j0=0; j0<KK; j0++) x[j0]=0.0;
for (i0=0; i0<KK; i0++) {
for (i1=0; i1<KK; i1++) {
for (i2=0; i2<KK; i2++) {
  j0=ascrule[i0][i1][i2]-'0';
  x[j0]+=a[i0]*a[i1]*a[i2];
  }}}
}

/* iterate the 1-block parameters to find self-consistent values */
/* graph the iterative steps in bar-chart form			 */
/* ll - initial line	*/
/* mm - length of line	*/
/* nn - number of lines	*/

oneblfreq(ll,mm,nn) int ll, mm, nn; {
int    ii, i, l, m, n;
double op[KK], np[KK];
double d, e, f, s;

m=0;
f=(double)mm;
s=1.0/((double)(KK));
n=(int)(f*s);
videodot(ll,m++,3);
for (i=0; i<KK; i++) {op[i]=s; for (l=0; l<n; l++) videodot(ll,m++,i);}

for (ii=1; ii<=nn; ii++) {
  e=0.0; m=0;
  onebl(np,op);
  videodot(ll+ii,m++,3);
  for (i=0; i<KK; i++) {
    n=(int)(f*np[i]); if (n>0) for (l=0; l<n; l++) videodot(ll+ii,m++,i);
    d=op[i]-np[i];
    e+=d*d;
    op[i]=np[i];
    }
  videodot(ll+ii,m++,3);
  if (op[0]<=0.001) break;
  if (op[1]<=0.001) break;
  if (e<=0.0000001) break;
  if (kbdst()) {kbdin(); break;}
  }
videocursor(0,BROW+8,0);
printf("(1-block) "); 
for (i=0; i<KK; i++) printf("%2d:%5.3f ",i,op[i]);
videotrdot(TROW,TCOL,op,l);
}

/* evaluate the two-block probabilities after one generation */
twobl(x,a) double x[KK][KK], a[KK][KK]; {
int    i0, i1, i2, i3;
int    j0, j1;
double w, b[KK];

for (j0=0; j0<KK; j0++) {for (j1=0; j1<KK; j1++) x[j0][j1]=0.0;}

for (j0=0; j0<KK; j0++) {b[j0]=0.0; for (j1=0; j1<KK; j1++) b[j0]+=a[j0][j1];}

for (i0=0; i0<KK; i0++) {
for (i1=0; i1<KK; i1++) {
for (i2=0; i2<KK; i2++) {
for (i3=0; i3<KK; i3++) {
  j0=ascrule[i0][i1][i2]-'0';
  j1=ascrule[i1][i2][i3]-'0';
  w=a[i0][i1]*a[i1][i2]*a[i2][i3];
  if (w!=0.0) w/=b[i1]*b[i2];
  x[j0][j1]+=w;
  }}}}
}

/* iterate the 2-block parameters to find self-consistent values */
/* graph the iterative steps in bar-chart form			 */
/* ll - initial line	*/
/* mm - length of line	*/
/* nn - number of lines	*/
twoblfreq(ll,mm,nn) int ll, mm, nn; {
int    ii, i, j, l, m, n;
double op[KK][KK], np[KK][KK];
double b[KK], d, e, f, s;

m=0;
f=(double)mm;
s=1.0/((double)(KK*KK));
n=(int)(f*s);
videodot(ll,m++,3);
for (i=0; i<KK; i++) {
for (j=0; j<KK; j++) {
  op[i][j]=s;
  for (l=0; l<n; l++) videodot(ll,m++,j);
  }}
videodot(ll,m++,3);

for (ii=1; ii<=nn; ii++) {
  e=0.0; m=0;
  twobl(np,op);
  videodot(ll+ii,m++,3);
  for (i=0; i<KK; i++) {
  for (j=0; j<KK; j++) {
    n=(int)(f*np[i][j]);
    if (n>0) for (l=0; l<n; l++) videodot(ll+ii,m++,j);
    d=op[i][j]-np[i][j];
    e+=d<0.0?-d:d;
    op[i][j]=np[i][j];
    }}
  videodot(ll+ii,m++,3);
for (i=0; i<KK; i++) {b[i]=0.0; for (j=0; j<KK; j++) b[i]+=op[i][j];}
videotrdot(TROW,TCOL,b,1);
  if (e<=0.0001) break;
  if (kbdst()) {kbdin(); break;}
  }
videocursor(0,BROW+8,0);
printf("(2-block) "); 
for (i=0; i<KK; i++) printf("%2d:%5.3f ",i,b[i]);
videotrdot(TROW,TCOL,b,3);
}

/* evaluate the three-block probabilities after one generation */
thrbl(x,a) double x[KK][KK][KK], a[KK][KK][KK]; {
int    i0, i1, i2, i3, i4;
int    j0, j1, j2;
double w, b[KK][KK];

for (j0=0; j0<KK; j0++) {
for (j1=0; j1<KK; j1++) {
  for (j2=0; j2<KK; j2++) x[j0][j1][j2]=0.0;
  b[j0][j1]=0.0;
  for (j2=0; j2<KK; j2++) b[j0][j1]+=a[j0][j1][j2];  
  }}

for (i0=0; i0<KK; i0++) {
for (i1=0; i1<KK; i1++) {
for (i2=0; i2<KK; i2++) {
for (i3=0; i3<KK; i3++) {
for (i4=0; i4<KK; i4++) {
  j0=ascrule[i0][i1][i2]-'0';
  j1=ascrule[i1][i2][i3]-'0';
  j2=ascrule[i2][i3][i4]-'0';
  w=a[i0][i1][i2]*a[i1][i2][i3]*a[i2][i3][i4];
  if (w!=0.0) w/=b[i1][i2]*b[i2][i3];
  x[j0][j1][j2]+=w;
  }}}}}
}

/* iterate the 3-block parameters to find self-consistent values */
/* ll - initial line	*/
/* mm - length of line	*/
/* nn - number of lines	*/
thrblfreq(ll,mm,nn) int ll, mm, nn; {
int    ii, i, j, k, l, m, n;
double op[KK][KK][KK], np[KK][KK][KK];
double b[KK], bb[KK][KK], d, e, f, s;

m=0;
f=(double)mm;
s=1.0/((double)(KK*KK*KK));
n=(int)(f*s);
videodot(ll,m++,3);
for (i=0; i<KK; i++) {
for (j=0; j<KK; j++) {
for (k=0; k<KK; k++) {
  op[i][j][k]=s;
  for (l=0; l<n; l++) videodot(ll,m++,k);
  }}}
videodot(ll,m++,3);

for (ii=1; ii<=nn; ii++) {
  e=0.0; m=0;
  thrbl(np,op);
  videodot(ll+ii,m++,3);
  for (i=0; i<KK; i++) {
  for (j=0; j<KK; j++) {
  for (k=0; k<KK; k++) {
    n=(int)(f*np[i][j][k]);
    if (n>0) for (l=0; l<n; l++) videodot(ll+ii,m++,k);
    d=op[i][j][k]-np[i][j][k];
    e+=d<0.0?-d:d;
    op[i][j][k]=np[i][j][k];
    }}}
  videodot(ll+ii,m++,3);
  if (e<=0.0001) break;
  if (kbdst()) {kbdin(); break;}
  }

for (i=0; i<KK; i++) {
  b[i]=0.0;
  for (j=0; j<KK; j++) {
  for (k=0; k<KK; k++) {
    b[i]+=op[i][j][k];
    }} }
videocursor(0,BROW+8,0);
printf("(3-block) "); 
for (i=0; i<KK; i++) printf("%2d:%5.3f ",i,b[i]);
videotrdot(TROW,TCOL,op,l);

for (i=0; i<KK; i++) {
for (j=0; j<KK; j++) {
  bb[i][j]=0.0;
  for (k=0; k<KK; k++) {
    bb[i][j]+=op[i][j][k];
    }} }

/*videocursor(0,BROW+9,0);*/
/*for (i=0; i<KK; i++) for (j=0; j<KK; j++)*/ 
/*  printf("%1d%1d:%5.3f ",i,j,bb[i][j]);*/
/*videodot(199-(int)(100.0*bb[1][1]),BORG+(int)(100.0*b[1]),2);*/
}

/* evaluate the four-block probabilities after one generation */
foubl(x,a) double x[KK][KK][KK][KK], a[KK][KK][KK][KK]; {
int    i0, i1, i2, i3, i4, i5;
int    j0, j1, j2, j3;
double w, b[KK][KK][KK];

for (j0=0; j0<KK; j0++) {
for (j1=0; j1<KK; j1++) {
for (j2=0; j2<KK; j2++) {
  for (j3=0; j3<KK; j3++) x[j0][j1][j2][j3]=0.0;
  b[j0][j1][j2]=0.0;
  for (j3=0; j3<KK; j3++) b[j0][j1][j2]+=a[j0][j1][j2][j3];  
  }}}

for (i0=0; i0<KK; i0++) {
for (i1=0; i1<KK; i1++) {
for (i2=0; i2<KK; i2++) {
for (i3=0; i3<KK; i3++) {
for (i4=0; i4<KK; i4++) {
for (i5=0; i5<KK; i5++) {
  j0=ascrule[i0][i1][i2]-'0';
  j1=ascrule[i1][i2][i3]-'0';
  j2=ascrule[i2][i3][i4]-'0';
  j3=ascrule[i3][i4][i5]-'0';
  w=a[i0][i1][i2][i3]*a[i1][i2][i3][i4]*a[i2][i3][i4][i5];
  if (w!=0.0) w/=b[i1][i2][i3]*b[i2][i3][i4];
  x[j0][j1][j2][j3]+=w;
  }}}}}}
}

/* iterate the 4-block parameters to find self-consistent values */
/* ll - initial line	*/
/* mm - length of line	*/
/* nn - number of lines	*/
foublfreq(ll,mm,nn) int ll, mm, nn; {
int    ii, i0, i1, i2, i3, l, m, n;
double op[KK][KK][KK][KK], np[KK][KK][KK][KK];
double b[KK], bb[KK][KK], d, e, f, s;

m=0;
f=(double)mm;
s=1.0/((double)(KK*KK*KK*KK));
n=(int)(f*s);
videodot(ll,m++,3);
for (i0=0; i0<KK; i0++) {
for (i1=0; i1<KK; i1++) {
for (i2=0; i2<KK; i2++) {
for (i3=0; i3<KK; i3++) {
  op[i0][i1][i2][i3]=s;
  for (l=0; l<n; l++) videodot(ll,m++,i3);
  }}}}
videodot(ll,m++,3);

for (ii=1; ii<=nn; ii++) {
  e=0.0; m=0;
  foubl(np,op);
  videodot(ll+ii,m++,3);
  for (i0=0; i0<KK; i0++) {
  for (i1=0; i1<KK; i1++) {
  for (i2=0; i2<KK; i2++) {
  for (i3=0; i3<KK; i3++) {
    n=(int)(f*np[i0][i1][i2][i3]);
    if (n>0) for (l=0; l<n; l++) videodot(ll+ii,m++,i3);
    d=op[i0][i1][i2][i3]-np[i0][i1][i2][i3];
    e+=d<0.0?-d:d;
    op[i0][i1][i2][i3]=np[i0][i1][i2][i3];
    }}}}
  videodot(ll+ii,m++,3);
  if (e<=0.0001) break;
  if (kbdst()) {kbdin(); break;}
  }

for (i0=0; i0<KK; i0++) {
  b[i0]=0.0;
  for (i1=0; i1<KK; i1++) {
  for (i2=0; i2<KK; i2++) {
  for (i3=0; i3<KK; i3++) {
    b[i0]+=op[i0][i1][i2][i3];
    }}} }

videocursor(0,BROW+8,0);
printf("(4-block) "); 
for (i0=0; i0<KK; i0++) printf("%2d:%5.3f ",i0,b[i0]);
videotrdot(TROW,TCOL,op,l);

for (i0=0; i0<KK; i0++) {
for (i1=0; i1<KK; i1++) {
  bb[i0][i1]=0.0;
  for (i2=0; i2<KK; i2++) {
  for (i3=0; i3<KK; i3++) {
    bb[i0][i1]+=op[i0][i1][i2][i3];
    }} }}

/*videocursor(0,BROW+9,0);*/
/*for (i0=0; i0<KK; i0++) for (i1=0; i1<KK; i1++)*/ 
/*  printf("%1d%1d:%5.3f ",i0,i1,bb[i0][i1]);*/
/*videodot(199-(int)(100.0*bb[1][1]),BORG+(int)(100.0*b[1]),2);*/
}

/* evaluate the five-block probabilities after one generation */
fivbl(x,a) double x[KK][KK][KK][KK][KK], a[KK][KK][KK][KK][KK]; {
int    i0, i1, i2, i3, i4, i5, i6;
int    j0, j1, j2, j3, j4;
double w, b[KK][KK][KK][KK];

for (j0=0; j0<KK; j0++) {
for (j1=0; j1<KK; j1++) {
for (j2=0; j2<KK; j2++) {
for (j3=0; j3<KK; j3++) {
  for (j4=0; j4<KK; j4++) x[j0][j1][j2][j3][j4]=0.0;
  b[j0][j1][j2][j3]=0.0;
  for (j4=0; j4<KK; j4++) b[j0][j1][j2][j3]+=a[j0][j1][j2][j3][j4];  
  }}}}

for (i0=0; i0<KK; i0++) {
for (i1=0; i1<KK; i1++) {
for (i2=0; i2<KK; i2++) {
for (i3=0; i3<KK; i3++) {
for (i4=0; i4<KK; i4++) {
for (i5=0; i5<KK; i5++) {
for (i6=0; i6<KK; i6++) {
  j0=ascrule[i0][i1][i2]-'0';
  j1=ascrule[i1][i2][i3]-'0';
  j2=ascrule[i2][i3][i4]-'0';
  j3=ascrule[i3][i4][i5]-'0';
  j4=ascrule[i4][i5][i6]-'0';
  w=a[i0][i1][i2][i3][i4]*a[i1][i2][i3][i4][i5]*a[i2][i3][i4][i5][i6];
  if (w!=0.0) w/=b[i1][i2][i3][i4]*b[i2][i3][i4][i5];
  x[j0][j1][j2][j3][j4]+=w;
  }}}}}}}
}

/* iterate the 5-block parameters to find self-consistent values */
/* ll - initial line	*/
/* mm - length of line	*/
/* nn - number of lines	*/
fivblfreq(ll,mm,nn) int ll, mm, nn; {
int    ii, i0, i1, i2, i3, i4, l, m, n;
double op[KK][KK][KK][KK][KK], np[KK][KK][KK][KK][KK];
double b[KK], bb[KK][KK], d, e, f, s;

m=0;
f=(double)mm;
s=1.0/((double)(KK*KK*KK*KK*KK));
n=(int)(f*s);
videodot(ll,m++,3);
for (i0=0; i0<KK; i0++) {
for (i1=0; i1<KK; i1++) {
for (i2=0; i2<KK; i2++) {
for (i3=0; i3<KK; i3++) {
for (i4=0; i4<KK; i4++) {
  op[i0][i1][i2][i3][i4]=s;
  for (l=0; l<n; l++) videodot(ll,m++,i4);
  }}}}}
videodot(ll,m++,3);

for (ii=1; ii<=nn; ii++) {
  e=0.0; m=0;
  fivbl(np,op);
  videodot(ll+ii,m++,3);
  for (i0=0; i0<KK; i0++) {
  for (i1=0; i1<KK; i1++) {
  for (i2=0; i2<KK; i2++) {
  for (i3=0; i3<KK; i3++) {
  for (i4=0; i4<KK; i4++) {
    n=(int)(f*np[i0][i1][i2][i3][i4]);
    if (n>0) for (l=0; l<n; l++) videodot(ll+ii,m++,i4);
    d=op[i0][i1][i2][i3][i4]-np[i0][i1][i2][i3][i4];
    e+=d<0.0?-d:d;
    op[i0][i1][i2][i3][i4]=np[i0][i1][i2][i3][i4];
    }}}}}
  videodot(ll+ii,m++,3);
  if (e<=0.0001) break;
  if (kbdst()) {kbdin(); break;}
  }
for (i0=0; i0<KK; i0++) {
  b[i0]=0.0;
  for (i1=0; i1<KK; i1++) {
  for (i2=0; i2<KK; i2++) {
  for (i3=0; i3<KK; i3++) {
  for (i4=0; i4<KK; i4++) {
    b[i0]+=op[i0][i1][i2][i3][i4];
    }}}} }
videocursor(0,BROW+8,0);
printf("(5-block) "); 
for (i0=0; i0<KK; i0++) printf("%2d:%5.3f ",i0,b[i0]);
videotrdot(TROW,TCOL,op,l);

for (i0=0; i0<KK; i0++) {
for (i1=0; i1<KK; i1++) {
  bb[i0][i1]=0.0;
  for (i2=0; i2<KK; i2++) {
  for (i3=0; i3<KK; i3++) {
  for (i4=0; i4<KK; i4++) {
    bb[i0][i1]+=op[i0][i1][i2][i3][i4];
    }} }}}

/*videocursor(0,BROW+9,0);*/
/*for (i0=0; i0<KK; i0++) for (i1=0; i1<KK; i1++) */
/*  printf("%1d%1d:%5.3f ",i0,i1,bb[i0][i1]);*/
/*videodot(199-(int)(100.0*bb[1][1]),BORG+(int)(100.0*b[1]),2);*/
}

/* end */
